/* -*- C++ -*-
 * mu_filterqueue.C
 * @(#) class for the modified queue needed by the probabilistic verifier
 *
 * Copyright (C) 2001 - 2003 by Giuseppe Della Penna              
 * Universita' degli Studi di L'Aquila
 * 
 * Read the file "license" distributed with these sources, or call
 * Murphi with the -l switch for additional information.
 * 
 */

/* 
 * Original Author: G. Della Penna
 * Created: 01 Oct 2001
 * Released: 30 Jan 2003
 *       
 */

/*
 * 
 */
 

#include "mu_probqueue.h"
#include <math.h>

#define min(a,b) (((a)<(b))?(a):(b))
#define max(a,b) (((a)>(b))?(a):(b))

#define DBG 0
#ifdef MAX_HT_CHAIN_LENGTH
#undef MAX_HT_CHAIN_LENGTH
#endif
#define MAX_HT_CHAIN_LENGTH 20
#define SPLITFILE_LEN (1024 * 0x100000L)


#define SFQ_PAGING_FILE_TOP "SFQ_PAGING_FILE_1"
#define SFQ_PAGING_FILE_BOTTOM "SFQ_PAGING_FILE_2"
#define SSQ_PAGING_FILE_TOP "SSQ_PAGING_FILE_1"
#define SSQ_PAGING_FILE_BOTTOM "SSQ_PAGING_FILE_2"
#define SFQ_HISTORY_FILE "SFQ_HISTORY_FILE"

char *make_unique_filename(char *basepathname)
{
  static char filenamebuffer[500];

  //Maybe I can add a random seed...
  sprintf(filenamebuffer,"%s_%s.swp",
      basepathname,PROTOCOL_NAME);
  
  return filenamebuffer;
}


state_probability_queue::state_probability_queue( unsigned long mas )
: max_active_states((unsigned long)(mas * gPercentActiveStates))
{
  
  stateArray = new state_and_probability_pair[ max_active_states];

  hashtable_size = NextPrime((unsigned long)mas);

  table = new state_and_probability_pair[hashtable_size];

  Full = new dynBitVec(hashtable_size);

  paging_file_top = new splitFile(SPLITFILE_LEN, false);
  paging_file_bottom = new splitFile(SPLITFILE_LEN, false);

  if ( !(paging_file_top->open(make_unique_filename(SFQ_PAGING_FILE_TOP),"w+b"))) {
      Error.Notrace( "Internal: Error creating top paging file for s_f_q.");
  }

  if ( !(paging_file_bottom->open(make_unique_filename(SFQ_PAGING_FILE_BOTTOM),"w+b"))) {
      Error.Notrace( "Internal: Error creating bottom paging file.");
  }
  
  num_elts_head = num_elts_tail = 0;

  head_begin = 0;
  tail_begin = max_active_states/2;

  head_size = max_active_states/2;
  tail_size = max_active_states - head_size;

  global_front=global_rear=front=rear=0;
}


state_probability_queue::~state_probability_queue()
{
  delete[] stateArray; 
  delete[] table; 
  delete Full;

  paging_file_top->close(); //rmtmp();
  paging_file_bottom->close(); //rmtmp();

  delete paging_file_top;
  delete paging_file_bottom;
}

bool state_probability_queue::isempty( void ) 
{ 
#if (DBG > 0) 
fprintf(stderr, 
"isempty(): num_elts_head=%d num_elts_tail=%d global_rear=%d global_front=%d",num_elts_head, num_elts_tail, global_rear, global_front);
#endif
    return num_elts_head + num_elts_tail + global_rear + global_front + hashtable_num_elts == 0; 
}

int 
state_probability_queue::BytesForOneState( void ) {
  return sizeof(state_and_probability_pair);  
}

int 
state_probability_queue::BytesForOneState_hashtable( void ) {
  return sizeof(state_and_probability_pair);  
}



void
state_probability_queue::Print( void )
{
  unsigned long i;
    
  for( i = 0; i < num_elts_head; i++ )
    {
      // convert to print in unsigned long format?
      cout << "State " << i << " [" << head_begin+i << "]:\n";
      stateArray[ head_begin+i ]._state.print();
    }

  for( i = 0; i < num_elts_tail; i++ )
    {
      // convert to print in unsigned long format?
      cout << "State " << i << " [" << tail_begin+i << "]:\n";
      stateArray[ tail_begin+i ]._state.print();
    }
}


bool 
state_probability_queue::enqueue( state*& e, double p )
{
  if (hashtable_num_elts >= hashtable_size) CheckTable();

   state_and_probability_pair pair(e,p);
   state_and_probability_pair *ppair = &pair;
   

   if (!hashtable_lookup(ppair)) {
      ppair->_probability = p;
      e = &(ppair->_state);
      return true;
   } else {
      ppair->_probability+=p;
      e = &(ppair->_state);
      return false;
   }
} 


void
state_probability_queue::enqueue_commit( state_and_probability_pair* e )
{

  if( num_elts_tail >= tail_size )
   { //memory full: reclaim more space by swapping out the current queue
      ReclaimFreeSpace();
   }

    /*
      at this point, ReclaimFreeSpace has obtained new space in the queue and
      set the offsets (front, rear, ...) accordingly; so we proceed with
      the insertion without checking...
    */
    
    
    stateArray[ tail_begin + rear ] = *e;
    rear++;
    num_elts_tail++;
} 

state* 
state_probability_queue::dequeue( double &prob )
{ 
#if (DBG > 0) 
fprintf(stderr, "dequeue(): num_elts_head=%d",num_elts_head );
#endif

  state* retval;

  if( num_elts_head <= 0 )
    {
      QueueEmptyFault();
    }
  
  retval = &(stateArray[head_begin + front]._state);
  prob = stateArray[head_begin + front]._probability;

  front++;
  num_elts_head--;


#if (DBG > 0) 
fprintf(stderr, "dequeue(): prob=%f",prob );
#endif


  return retval;
}


state* 
state_probability_queue::top( double &prob )
{
  if ( num_elts_head <= 0 )
    {
      QueueEmptyFault();
    }

  prob = stateArray[head_begin + front]._probability;
  return &(stateArray[head_begin + front]._state);
}


void state_probability_queue::ReclaimFreeSpace()
{
#if (DBG > 0) 
   fprintf(stderr, "\nReclaimFreeSpace(sfq): writing %d states (%d bytes each) to disk", tail_size,sizeof(state_and_probability_pair));
#endif

    global_rear+= paging_file_bottom->write(&stateArray[tail_begin],sizeof(state_and_probability_pair),tail_size);
    
    num_elts_tail = 0;
    rear = 0;
}


void state_probability_queue::QueueEmptyFault()
{
#if (DBG > 0) 
    fprintf(stderr, "\nQueueEmptyFault(sfq): ");
#endif
    //make sure we do not read garbage after the states
    size_t read = paging_file_top->read(&stateArray[head_begin],sizeof(state_and_probability_pair),min(global_front,head_size));

    if (read>0) { //ok, some states are swapped in

        num_elts_head = read;
        global_front -= read;
#if (DBG > 0) 
	fprintf(stderr, " read %d states from bottom",read);
#endif

    } else if (global_rear>0) { //paging_file_top is empty, but paging_file_bottom is not

    splitFile *fswap;

    //swap files
#if (DBG > 0) 
    fprintf(stderr, " swapping top with bottom");
#endif
    fswap= paging_file_top;   
        paging_file_top = paging_file_bottom;

    paging_file_bottom = fswap;

        paging_file_top->seek(0,SEEK_SET); //move to the beginning of the queue
        paging_file_bottom->seek(0,SEEK_SET); //reset bottom queue

        global_front = global_rear;
        global_rear = 0; //bottom file is empty
        
        //now bottom entries are top entries and bottom file is empty. Reload a block!
        size_t read = paging_file_top->read(&stateArray[head_begin],sizeof(state_and_probability_pair),min(global_front,head_size));
#if (DBG > 0) 
	fprintf(stderr, " read %d states from disk",read);
#endif

        num_elts_head = read;
        global_front -= read;

    } else if (num_elts_tail >0){  //paging_file_top AND paging_file_bottom are empty

#if (DBG > 0) 
	fprintf(stderr, " using memory tail...");
#endif

        /* the disk queue is ended. this means that the only states we have
           to explore are the ones in the current tail window */
        unsigned long swap = tail_begin;
        tail_begin = head_begin;
        head_begin = swap;

        swap=tail_size;
        tail_size = head_size;
        head_size = swap;

        num_elts_head = num_elts_tail;
        num_elts_tail = 0;
        rear =0;
    } else { //no more states in both swap files, and the memory is empty: why do we call again?
        Error.Notrace( "Internal: Attempt to read an empty state queue.");
    }
    
    front = 0;
}


bool 
state_probability_queue::hashtable_lookup(state_and_probability_pair *&in)
{
  unsigned long key = in->_state.hashkey();
  unsigned long h1 = key % hashtable_size;
  unsigned long h2 = 1 + key % ( hashtable_size - 1 );
  unsigned long h = h1;

  unsigned long probe = 0;
// no hash compaction, uses double hashing

  bool empty, equal= FALSE;

  while ( !(empty = (Full->get(h) == 0)) &&
	  !(equal = ( in->_state == table[h]._state )) &&
           ( probe < MAX_HT_CHAIN_LENGTH ) )
    {
      h = (h1 + probe * h2) % hashtable_size;   // double hashing
      probe++;
    }
  if (empty)  /* Go ahead and insert the element. */
    {

      table[h] = *in;
      in = &table[h];
      Full->set(h);
      hashtable_num_elts++;

      return FALSE;
    }
  else if (equal)
    {
      in = &table[h];

      return TRUE;
    }
  else 
    {
#if (DBG > 0) 
      fprintf(stderr, "hashtable_lookup: table is full,clear it!\n");
#endif
      //table is full,clear it!
      CheckTable();
      //the table now is empty: use the first slot available for this state
      h = h1 % hashtable_size;

      table[h] = *in;
      in = &table[h];
      Full->set(h);
      hashtable_num_elts++;

      return FALSE;
    }

  return FALSE;
}


void state_probability_queue::CheckTable()
{
//1. move all the table states in the queue

   unsigned long moved=0;
   unsigned long i;

#if (DBG > 0) 
fprintf(stderr, "CheckTable(): begin state migration...\n");
#endif

   for (i=0 ; i< hashtable_size; i++) {
	if (Full->get(i)==1) {
		enqueue_commit(&table[i]);
		moved++; //for dbg only
	}
   }

#if (DBG > 0) 
fprintf(stderr, "CheckTable(): moved %lu states to the queue\n",moved);
#endif

//2. clear the table

   Full->clearall();
   hashtable_num_elts=0;

}



