/* -*- C++ -*-
 * mu_filterqueue.h
 * @(#) class for the modified queue needed by the probabilistic verifier
 *
 * Copyright (C) 2001 - 2003 by Giuseppe Della Penna              
 * Universita' degli Studi di L'Aquila
 * 
 * Read the file "license" distributed with these sources, or call
 * Murphi with the -l switch for additional information.
 * 
 */

/* 
 * Original Author: G. Della Penna
 * Created: 01 Oct 2001
 * Released: 30 Jan 2003
 *       
 */

/*
 * 
 */
 
#ifndef _FILTERQUEUE_
#define _FILTERQUEUE_


struct state_and_probability_pair {
    state _state;
    double _probability;

    state_and_probability_pair (state* s, double p) 
    {_state = *s; _probability=p;}

    state_and_probability_pair () { };
};


/****************************************
  The state queue.
 ****************************************/
class state_probability_queue
{
#ifdef HASHC
  typedef unsigned long Unsigned32; 
#endif


protected:
  state_and_probability_pair* stateArray;                     /* The actual array. */
  state_and_probability_pair *table;

  const unsigned long max_active_states;  /* max size of queue */
  unsigned long front;                    /* index of first active state. */
  unsigned long rear;                     /* index of next free slot. */

  unsigned long global_rear,global_front;
  unsigned long num_elts_head,num_elts_tail;
  unsigned long head_begin,tail_begin;
  unsigned long head_size,tail_size;

  splitFile *paging_file_top, *paging_file_bottom;

  bool hashtable_lookup(state_and_probability_pair* &in);
  void enqueue_commit( state_and_probability_pair* e );

  dynBitVec *Full;

  unsigned long hashtable_size;
  unsigned long hashtable_num_elts;

  randomGen random;

public:
  // initializers
  state_probability_queue( unsigned long mas );

  // destructor
  virtual ~state_probability_queue();
  
  // information interface
  inline unsigned long MaxElts( void ) { return max_active_states; } 
  unsigned long NumElts( void ) { return num_elts_head + num_elts_tail + global_rear + global_front + hashtable_num_elts; }
  inline bool isempty( void );
    
  inline static int BytesForOneState( void ); 
  inline static int BytesForOneState_hashtable( void ); 
  
  // storing and removing elements
  virtual bool enqueue( state* &e, double p );
  virtual state* dequeue( double &prob );
  virtual state* top( double &prob );
  
  virtual void ReclaimFreeSpace();
  virtual void QueueEmptyFault();

  inline void CheckTable();

  virtual unsigned NextRuleToTry()   // Uli: unsigned short -> unsigned
  {
    Error.Notrace("Internal: Getting next rule to try from a state queue instead of a state stack.");
    return 0;
  }
  virtual void NextRuleToTry(unsigned r)
  {
    Error.Notrace("Internal: Setting next rule to try from a state queue instead of a state stack.");
  }

  // printing routine
  void Print( void );
  virtual void print_capacity( void )
  {
    cout << "\t* Capacity of queue for breadth-first search:\n"
     << "\t  * " << max_active_states << " states in the queue.\n"
     << "\t  * " << hashtable_size << " states in the cache.\n"
     << "\t   * Change the constant gPercentActiveStates in mu_prolog.inc\n"
         << "\t     to increase this, if necessary.\n"; 
  }
};

#endif
